import click
from snark.log import logger
from snark import config
from snark.client.pod_control import PodControlClient
from snark.cli.connect import get_pod_key_path, install_pod_key
from snark.cli.utils import get_proxy_command

from subprocess import call
import os
import subprocess

@click.command()
@click.argument('pod_id', default='my_pod')
@click.option('--local_path', '-l', default=None, type=str, help='Path to file on local machine.')
@click.option('--remote_path', '-r', default=None, type=str, help='Path to file on the Pod.')
def push(pod_id, local_path, remote_path):
    """ Transfer file to the pod """
    if not pod_id:
        pod_id = click.prompt('Pod ID', type=str)
    pod_id = pod_id.strip()


    if not local_path:
        local_path = click.prompt('Local path', type=str)
    local_path = local_path.strip()


    if not remote_path:
        remote_path = click.prompt('Remote path', type=str)
    remote_path = remote_path.strip()

    proxy, address, username, private_key, port = PodControlClient().get_connect_info(pod_id)
    install_pod_key(username, pod_id, private_key)
    key_file_path = get_pod_key_path(username, pod_id)
    ssh_proxy = get_proxy_command(proxy)

    command = 'scp -i "{}" -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null\
                    {} -P {} {} {}@{}:{} '.format(key_file_path, ssh_proxy,
                                              port, local_path, username, address, remote_path)
    os.system(command)

@click.command()
@click.argument('pod_id', default='my_pod')
@click.option('--remote_path', '-r', default=None, type=str, help='Path to file on the Pod.')
@click.option('--local_path', '-l', default=None, type=str, help='Path to file on local machine.')
def pull(pod_id, remote_path, local_path):
    """ Transfer file from the pod to the localhost """
    if not pod_id:
        pod_id = click.prompt('Pod ID', type=str)
    pod_id = pod_id.strip()


    if not local_path:
        local_path = click.prompt('Local path', type=str)
    local_path = local_path.strip()


    if not remote_path:
        remote_path = click.prompt('Remote path', type=str)
    remote_path = remote_path.strip()

    proxy, address, username, private_key, port = PodControlClient().get_connect_info(pod_id)
    install_pod_key(username, pod_id, private_key)
    key_file_path = get_pod_key_path(username, pod_id)

    ssh_proxy = get_proxy_command(proxy)

    command = 'scp -i "{}" -oStrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null \
                {} -P {} {}@{}:{} {}'.format(key_file_path, ssh_proxy,
                                               port, username, address, remote_path, local_path)
    os.system(command)



import websocket
import time
import json
import requests
import uuid
import numpy as np
from datetime import datetime

from snark.token_manager import TokenManager
from snark.infer.message_processor import construct_infer_message, parse_data_message

class SnarkModel(object):
    """Snark Infer model"""

    def __init__(self, model_name):
        self.host = 'core.snark.ai'
        self.ws_port = 10004
        self.endpoint = "ws://{}:{}".format(self.host, self.ws_port)
        self.model_name = model_name
        self.input_shape = None
        self.output_shape = None
        self.model_info = {}
        self.task_id = uuid.uuid4()
        global snark_token

        self.token = TokenManager.get_token()
        self.ws = websocket.create_connection(self.endpoint)
        self.ws.settimeout(300)  # 300 secs timeout for receiving data
        self._connect()

    def __del__(self):
        try:
            self.ws.close()
        except:
            pass

    def _connect(self):
        # authentication message
        self.ws.send(json.dumps({"api_key": self.token,
                                 "payload": {
                                     "model": "/" + self.model_name,
                                     "task_id": str(self.task_id)
                                 }
                                 })
                     )
        reply = self.ws.recv()  # blocking call
        try:
            r = json.loads(reply)
        except Exception as e:
            raise Exception(
                "Failed to create a model connection. Server side error.")

        if (not "status" in r) or (r["status"] != "successful"):
            if not "description" in r:
                raise Exception(
                    "Failed to create a model connection. Server side error.")

            raise Exception(
                "Failed to create model connection. {}".format(r["description"]))
        else:
            self.model_info = r["model_info"]
            self.input_shape = tuple(json.loads(
                self.model_info["input_shape"]))
            self.output_shape = tuple(json.loads(
                self.model_info["output_shape"]))

    def infer(self, np_input, verbose=False):
        assert (np_input.shape == self.input_shape) or \
               (np_input[np.newaxis, ...].shape == self.input_shape)

        if (np_input[np.newaxis, ...].shape == self.input_shape):
            np_input = np_input[np.newaxis, ...]
        try:
            infer_message = construct_infer_message(np_input)
        except Exception as e:
            raise e

        self.ws.send(infer_message, opcode=websocket.ABNF.OPCODE_BINARY)

        if verbose:
            sent_time = datetime.now()
            print('[{}] Request sent...'.format(sent_time))
        try:
            result_raw = self.ws.recv()
        except Exception as e:
            raise Exception(
                "Failed to receive result from the server. Please try again.")


        if verbose:
            received_time = datetime.now()
            elapsed_time = received_time - sent_time
            elapsed_secs = elapsed_time.total_seconds()
            print("[{}] Reply received. Time elapsed: {} sec".format(
                received_time, elapsed_secs))

        result_np = parse_data_message(result_raw, self.output_shape)
        return result_np

def model(model_name):
    return SnarkModel(model_name)




import json

import snark
from snark import config
from snark.log import logger
from snark.client.base import SnarkHttpClient
from snark.exceptions import SnarkException

class HyperControlClient(SnarkHttpClient):
    """
    Controlling Snark Pods through rest api
    """
    def __init__(self):
        super(HyperControlClient, self).__init__()

    def upload(self, descriptor):
        files = {"descriptor": descriptor}

        response = self.request(
            'POST', config.UP_DESCRIPTOR_SUFFIX,
            endpoint=config.SNARK_HYPER_ENDPOINT,
            files=files
        )
        return response.json()

    def down(self, descriptor):
        files = {"descriptor": descriptor}
        response = self.request(
            'POST', config.DOWN_DESCRIPTOR_SUFFIX,
            endpoint=config.SNARK_HYPER_ENDPOINT,
            files=files
        )
        return response.json()

    def list(self):
        response = self.request(
            'GET', config.LIST_EXPERIMENTS_SUFFIX,
            endpoint=config.SNARK_HYPER_ENDPOINT
        )
        return response.json()

    def list_task(self, experiment_id):
        data = {"experiment_id": experiment_id}
        response = self.request(
            'GET', config.LIST_EXPERIMENT_SUFFIX,
            endpoint=config.SNARK_HYPER_ENDPOINT,
            json=data
        )
        return response.json()



import click
import datetime
from tabulate import tabulate
from random  import randint

from snark.log import logger
from snark import config
from snark.client.pod_control import PodControlClient
from snark.cli.connect import connect
from snark.cli.auth import login
from snark.token_manager import TokenManager


@click.command()
@click.pass_context
def ls(ctx):
    """ List active pods"""
    logger.debug("Listing Pods.")
    if not TokenManager.is_authenticated():
        ctx.invoke(login)

    pods_dict = PodControlClient().list_active()
    pods_list = [
                  [p,
                      datetime.datetime.fromtimestamp(
                                  float(pods_dict[p]['start_time'])
                                      ).strftime('%Y-%m-%d %H:%M:%S'),
                   pods_dict[p]['GPU_spec']]
                for p in pods_dict]
    print (tabulate(pods_list, headers=['Name', 'Time Start', 'GPU spec']))

@click.command()
@click.argument('pod_id', default='')
@click.option('--pod_type', '-t', default='pytorch', type=click.Choice(config.POD_TYPES),
                help='Pod type')
@click.option('--gpu_count', '-g', type=int, default=1, help='Number of GPUs')
@click.option('--port_forwarding', '-L', default='', help='Port forwarding')
@click.option('--jupyter', '-j', flag_value=True, default=False, help='Jupyter running')
@click.option('--gpu_spec', '-s', type=click.Choice(config.GPU_TYPES),
                                  default=config.GPU_TYPES[0],
                                  help='GPU type')
@click.option('--docker_image', '-d', default=None, help='Dockerhub image name for custom pod')
@click.option('--gpu_count', '-g', type=int, default=1, help='Number of GPUs')
@click.option('--ssh/--no-ssh', '-j', default=True, help='Weather to connect to pod after start')
@click.pass_context
def start(ctx, pod_id, pod_type, gpu_count, gpu_spec, port_forwarding, jupyter, docker_image, ssh):
    """ Launch a Snark AI pod"""
    logger.info("Setting up the pod...")

    if not TokenManager.is_authenticated():
        ctx.invoke(login)

    if not pod_id:
        #pod_id = click.prompt('Pod ID', type=str)
        pod_id = "pod_{}".format(''.join(["%s" % randint(0, 9) for num in range(0, 5)]))
    pod_id = pod_id.strip()

    ## check if pod exists
    active_pods = PodControlClient().list_active()
    #FIXME Check if portforwarding is in correct format o/w will create the pod but will fail

    if pod_id not in active_pods:
        pod_type = pod_type.strip()

        PodControlClient().create_pod(pod_id, pod_type, docker_image)
        logger.debug("Successfully created pod {}.".format(pod_id))

        hardware_spec = {"CPU_count": 1, "GPU_count": gpu_count, "GPU_type": gpu_spec}

        timeout = config.DEFAULT_TIMEOUT
        if pod_type == 'custom':
            timeout = config.DEFAULT_TIMEOUT * 10

        PodControlClient().start_pod(pod_id, hardware_spec, timeout=timeout)
        logger.debug("Successfully started pod {}.".format(pod_id))
        if ssh:
            connect(pod_id, jupyter=jupyter, port_forwarding=port_forwarding)
    else:
        logger.info("Pod with ID `{}` is already running. Please use `snark attach {}` to reattach.".format(pod_id, pod_id))

def terminate(pod_id):
    """ Terminate the given pod"""
    logger.debug("Terminating pod.")
    pod_id = pod_id.strip()

    PodControlClient().terminate_pod(pod_id)
    logger.debug("Successfully terminated pod {}.".format(pod_id))


@click.command()
@click.argument('pod_id')
@click.pass_context
def stop(ctx, pod_id):
    """ Stop the given pod """
    logger.info("Stopping pod...")
    if not TokenManager.is_authenticated():
        ctx.invoke(login)
    if not pod_id:
        pod_id = click.prompt('Pod ID', type=str)
    pod_id = pod_id.strip()

    PodControlClient().stop_pod(pod_id)
    terminate(pod_id)
    logger.debug("Successfully stopped pod {}.".format(pod_id))



import numpy as np
import uuid
import json

def construct_infer_message(numpy_array):
    imgdata = numpy_array.ravel().tobytes()
    return imgdata

def parse_data_message(imgdata, dims):
    # through reasonable exception if message is of bad format
    np_array = np.fromstring(imgdata, dtype=np.float32).reshape(*dims)
    return np_array



import json

import snark
from snark import config
from snark.log import logger
from snark.client.base import SnarkHttpClient
from snark.exceptions import SnarkException

class PodControlClient(SnarkHttpClient):
    """
    Controlling Snark Pods through rest api
    """
    def __init__(self):
        super(PodControlClient, self).__init__()

    def create_pod(self, pod_id, pod_type, custom_docker_image):
        response = self.request('GET', config.CREATE_POD_REST_SUFFIX,
                                json={"pod_id": pod_id,
                                      "pod_type": pod_type,
                                      "custom_docker_image": custom_docker_image})

    def terminate_pod(self, pod_id):
        response = self.request('GET', config.TERMINATE_POD_REST_SUFFIX,
                                json={"pod_id": pod_id})

    def start_pod(self, pod_id, hardware_spec, timeout=config.DEFAULT_TIMEOUT):
        response = self.request('GET', config.START_POD_REST_SUFFIX,
                json={"pod_id": pod_id, "hardware_spec": json.dumps(hardware_spec)}, timeout=timeout)

    def stop_pod(self, pod_id):
        response = self.request('GET', config.STOP_POD_REST_SUFFIX,
                                json={"pod_id": pod_id})

    def get_connect_info(self, pod_id):
        response = self.request('GET', config.GET_CONNECT_INFO_REST_SUFFIX,
                                json={"pod_id": pod_id})
        response = response.json()
        return (response['is_proxy'],
               response['address'],
               response['username'],
               response['private_key'],
               response['port'])

    def list_active(self):
        response = self.request('GET', config.LIST_ACTIVE_PODS_REST_SUFFIX)
        active_pods_json = response.json()['active_pods']
        active_pods = json.loads(active_pods_json)
        return active_pods



import click
from snark.log import logger
from snark.client.hyper_control import HyperControlClient
import yaml
from os import walk
import pprint
from tabulate import tabulate
import json
import os

def get_all_files(path = "./"):
    f = []
    for (dirpath, dirnames, filenames) in walk(path):
        f.append((dirpath, dirnames, filenames))
        #break
    #print(f)
    return f

@click.command()
@click.argument('name_id', default='my_experiment')
@click.option('--file', '-f', default='snark.yml', help='YAML descriptor file for uploading')
@click.pass_context
def up(ctx, file, name_id):
   """ up """
   get_all_files()
   with open(file, 'r') as stream:
       data_loaded = yaml.load(stream)
  # print(data_loaded)
   descriptor = open(file,'rb')
   print("Creating Experiments...")
   pods_dict = HyperControlClient().upload(descriptor)
   print("Starting Cluster...")

@click.command()
@click.argument('name_id', default='my_experiment')
@click.option('--file', '-f', default='snark.yml', help='YAML descriptor file for uploading')
@click.pass_context
def down(ctx, file, name_id):
   """ up """
   get_all_files()
   with open(file, 'r') as stream:
       data_loaded = yaml.load(stream)
   descriptor = open(file,'rb')
   pods_dict = HyperControlClient().down(descriptor)

@click.command()
@click.argument('name_id', default='')
@click.pass_context
def ps(ctx, name_id):
   """ up """
   if name_id == "":
        experiments = HyperControlClient().list()

        experiments = [[e['ID'][:6], e['Name'], e['State'], e['IPs'], len(e['Tasks'])]
                 for e in experiments]
        print (tabulate(experiments, headers=['ID', 'Name', 'State', 'IPs', 'Tasks']))
   else:
       exp = HyperControlClient().list_task(name_id)
       tasks = exp['Tasks']
       task_list = [[t['TaskId'][:6], t['State'], t['AssignedNode']]
                   for t in tasks]
       print (tabulate(task_list, headers=['Name', 'State', 'Node']))


@click.command()
@click.argument('name_id', default='')
@click.pass_context
def logs(ctx, name_id):
   """ up """
   if name_id == '':
       print('please provide experiment id')
       exit()

   exp = HyperControlClient().list_task(name_id)
   tasks = exp['Tasks']
   for task in tasks:
       print(str(json.loads(task['Logs'][0])['result']).replace("\\r\\n", "\n"))



import os
from snark import config

def set_up_file(path):
    full_path = os.path.expanduser(path)
    dir_name  = os.path.dirname(full_path)
    if not os.path.exists(dir_name):
        try:
            os.makedirs(dir_name)
        except OSError as exc: # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise

    #create an empty file
    with open(path, 'a+'):
        pass

set_up_file(config.TOKEN_FILE_PATH)



import snark
from snark import config
from snark.log import logger
from snark.client.base import SnarkHttpClient
from snark.exceptions import SnarkException

class AuthClient(SnarkHttpClient):
    """
    Authentication through rest api
    """
    def __init__(self):
        super(AuthClient, self).__init__()

    def check_token(self, access_token):
        auth = "Bearer {}".format(access_token)
        response = self.request('GET', config.CHECK_TOKEN_REST_SUFFIX, headers={"Authorization": auth})

        try:
            response_dict = response.json()
            is_valid = response_dict["is_valid"]
        except Exception as e:
            logger.error("Exception occured while validating token: {}.".format(e))
            raise SnarkException("Error while validating the token. \
                                  Please try logging in using username ans password.")

        return is_valid

    def get_access_token(self, username, password):
        response = self.request('GET', config.GET_TOKEN_REST_SUFFIX,
                                json={'username': username,
                                      'password': password})
        try:
            token_dict = response.json()
            token = token_dict['access_token']
        except Exception as e:
            logger.error("Exception occured while getting token: {}.".format(e))
            raise SnarkException("Error while loggin in. \
                                  Please try logging in using access token.")
        return token








import click
import sys

from snark.cli.auth          import login, logout
from snark.cli.pod_control   import start, stop, ls, admin_stop, admin_ls_all
from snark.cli.data_transfer import push, pull
from snark.cli.connect       import attach
from snark.cli.hyper            import up, down, ps, logs
from snark import config
from snark.log import configure_logger

@click.group()
@click.option('-h', '--host', default='{}'.format(config.SNARK_REST_ENDPOINT), help='Snark server rest endpoint')
@click.option('-v', '--verbose', count=True, help='Devel debugging')
def cli(host, verbose):
    configure_logger(verbose)

def add_commands(cli):
    cli.add_command(login)
    cli.add_command(logout)
    cli.add_command(start)
    cli.add_command(attach)
    cli.add_command(stop)
    cli.add_command(admin_stop)
    cli.add_command(admin_ls_all)
    cli.add_command(ls)
    cli.add_command(push)
    cli.add_command(pull)
    cli.add_command(up)
    cli.add_command(down)
    cli.add_command(ps)
    cli.add_command(logs)

add_commands(cli)



import os

curr_path = os.path.dirname(os.path.abspath(__file__))

TOKEN_FILE_PATH = os.path.expanduser("~/.snarkai/token")
POD_KEY_DIR_PATH = os.path.expanduser("~/.snarkai/pod_keys")

SNARK_REST_ENDPOINT  = "https://controller.snark.ai"
SNARK_HYPER_ENDPOINT = "https://hyper.snark.ai:5000"

UP_DESCRIPTOR_SUFFIX            = "/api/v1/experiment/up"
DOWN_DESCRIPTOR_SUFFIX          = "/api/v1/experiment/down"
LIST_EXPERIMENTS_SUFFIX         = "/api/v1/experiment/all"
LIST_EXPERIMENT_SUFFIX          = "/api/v1/experiment/id"

GET_TOKEN_REST_SUFFIX     = "/api/v1/get_token"
CHECK_TOKEN_REST_SUFFIX   = "/api/v1/check_token"

CREATE_POD_REST_SUFFIX       = "/api/v1/pod/create"
TERMINATE_POD_REST_SUFFIX    = "/api/v1/pod/terminate"
START_POD_REST_SUFFIX        = "/api/v1/pod/start"
STOP_POD_REST_SUFFIX         = "/api/v1/pod/stop"
GET_CONNECT_INFO_REST_SUFFIX = "/api/v1/pod/connect_info"
LIST_ACTIVE_PODS_REST_SUFFIX = "/api/v1/pod/ls"
ADMIN_STOP_POD_REST_SUFFIX       = "/api/v1/admin/pod/stop"
ADMIN_TERMINATE_POD_REST_SUFFIX  = "/api/v1/admin/pod/terminate"
ADMIN_LS_ALL_REST_SUFFIX         = "/api/v1/admin/pod/ls_all"

POD_TYPES = ['pytorch', 'tensorflow', 'mxnet', 'caffe', 'fast.ai', 'custom']
GPU_TYPES = ['P106', '1080', '1070', 'Q4000']

DEFAULT_TIMEOUT = 170
